use bevy::prelude::*;

#[derive(Debug, PartialEq, Clone)]
pub enum Camera {
    TwoD(Entity),
    ThreeD(Entity),
}

impl Camera {
    pub fn is_3d(&self) -> bool {
        match self {
            Camera::TwoD(_) => false,
            Camera::ThreeD(_) => true,
        }
    }

    pub fn is_2d(&self) -> bool {
        !self.is_3d()
    }

    fn create_2d(cmd: &mut Commands) -> Self {
        Self::TwoD(cmd.spawn(Camera2dBundle::default()).id())
    }

    fn create_3d(cmd: &mut Commands, window_extent: Vec2) -> Self {
        Self::ThreeD(
            cmd.spawn(Camera3dBundle {
                transform: Transform::from_xyz(200.0, -600.0, 600.0)
                    .looking_at(window_extent.extend(0.0), Vec3::Z),
                ..default()
            })
            .id(),
        )
    }
}

#[derive(Resource, Debug, PartialEq, Clone)]
pub struct GlobalState {
    pub cursor_position: Vec2,
    pub window_extent: Vec2,
    pub polygon_started: bool,
    pub view_2d: bool,
    pub extrusion_height: f32,

    pub camera: Option<Camera>,
}

impl GlobalState {
    pub fn setup_2d_cam(&mut self, cmd: &mut Commands) {
        match &mut self.camera {
            Some(camera) => match camera {
                Camera::TwoD(_) => (),
                Camera::ThreeD(cam) => {
                    cmd.entity(*cam).despawn();

                    *camera = Camera::create_2d(cmd);
                }
            },
            None => self.camera = Some(Camera::create_2d(cmd)),
        }
    }

    pub fn setup_3d_cam(&mut self, cmd: &mut Commands) {
        match &mut self.camera {
            Some(camera) => match camera {
                Camera::TwoD(cam) => {
                    cmd.entity(*cam).despawn();

                    *camera = Camera::create_3d(cmd, self.window_extent);
                }
                Camera::ThreeD(_) => (),
            },
            None => self.camera = Some(Camera::create_3d(cmd, self.window_extent)),
        }
    }
}

impl Default for GlobalState {
    fn default() -> Self {
        Self {
            cursor_position: Vec2::default(),
            window_extent: Vec2::default(),
            polygon_started: false,
            view_2d: true,
            extrusion_height: 20.0,

            camera: None,
        }
    }
}

#[cfg(test)]
mod test {
    use bevy::{ecs::system::CommandQueue, prelude::*};

    use super::GlobalState;

    #[test]
    fn camera_conversion() {
        let world = World::new();
        let mut command_queue = CommandQueue::default();

        let mut cmd = Commands::new(&mut command_queue, &world);

        // default global state
        let mut global_state = GlobalState::default();

        // create 2d cam
        global_state.setup_2d_cam(&mut cmd);

        assert!(global_state.camera.is_some());
        assert!(global_state.camera.as_ref().unwrap().is_2d());

        // switch to 3d cam
        global_state.setup_3d_cam(&mut cmd);

        assert!(global_state.camera.is_some());
        assert!(global_state.camera.as_ref().unwrap().is_3d());

        // switch to 2d cam
        global_state.setup_2d_cam(&mut cmd);

        assert!(global_state.camera.is_some());
        assert!(global_state.camera.as_ref().unwrap().is_2d());
    }
}
